local compile_pllua_function = ...
local concat = table.concat
local tostring = tostring
local format = string.format
local match = string.match
local find = string.find
local sub = string.sub
local gsub = string.gsub
local insert = table.insert

function IS(x)
    if x ~= nil and x ~= NULL then
        return x
    else
        return nil
    end
end

function IFNULL(x, def)
    if x ~= nil and x ~= NULL then
        return x
    else
        return def
    end
end

function NOTNULL(x)
    return x ~= nil and x ~= NULL
end

function ISNULL(x)
    return x == NULL or x == nil
end

function map(by, func, obj, first_index)
    local function transform(newind, ...)
      if not newind then return nil end
      return newind, by(...)
    end
    return function(obj, ind)
       return transform(func(obj, ind))
    end, obj, first_index
end

local function next_by_name(tuple, field)
    local __mapping__ = tuple.__mapping__
    if __mapping__ and (__mapping__[field] ~= nil or field == nil) then
      local indx
      repeat
        field, indx = next(__mapping__, field)
      until type(field) == 'string' or field == nil
      if field ~= nil then
        return field, tuple[indx]
      end
    end
    repeat
      field, value = next(tuple, field)
    until (type(field) == 'string' and field ~= '__mapping__') or field == nil
    return field, value
end

local function flds(fields, t)
    for k, v in pairs(t) do
      if type(k) == 'string' and k ~= '__mapping__' then
        insert(fields, k)
      end
    end
  end
local function tuple_field_names(tup)
  local fields = {}
  flds(fields, tup)
  if tup.__mapping__ then flds(fields, tup.__mapping__) end
  return fields
end

spi = spi or {}
spi.next_by_name = next_by_name
spi.tuple_field_names = tuple_field_names

--[[ 
function spi.tuple_by_name(tup)
  return next_by_name, tup, nil
end
--]]
local function next_from_field_names(fld_names, field)
  field = fld_names[field or 1]
  if field == nil then return nil, nil end
  return field, fld_names[0][field]
end

function spi.tuple_by_name(tup)
  local field_names = tuple_field_names(tup)
  for i,field in ipairs(field_names) do
    field_names[field] = field_names[i + 1] -- chain field names
  end
  field_names[0] = tup
  return next_from_field_names, field_names, nil
end

spi.error = error
function spi.empty(e)
  return ISNULL(e) or #e==0
end

spi_resultset = spi_resultset or {}
spi_resultset.iterate = ipairs

timestamptz = timestamptz or {}
timestamp = timestamp or {}
do
  local to_timestamp = timestamptz.to_timestamp
  local now, statement, clock = timestamptz.now, timestamptz.statement, timestamptz.clock
  function timestamp.now()
    return to_timestamp(now())
  end
  function timestamp.statement()
    return to_timestamp(statement())
  end
  function timestamp.clock()
    return to_timestamp(clock())
  end
end


do
  interval = interval or {}
  local fmod, modf = math.fmod, math.modf
  function math.rint(v)
    if v > 0 then
      return (modf( v + 0.5))
    elseif v < 0 then
      return (modf( v - 0.5))
    else
      return v
    end
  end
  local rint = math.rint
  function math.rintn(v, n)
    local koef = 10 ^ (n or 0)
    return rint(v * koef)/koef
  end
  local rintn = math.rintn
  function interval.new(seconds, days, months)
    return {seconds = (seconds or 0), days = (days or 0), months = (months or 0)}
  end
  function interval.denormalize(int)
    local seconds = int.seconds or 0
    local days = int.days or 0
    local months = int.months or 0
    return {
      seconds = rintn(fmod(seconds, 60), 6),
      minutes = fmod(modf(seconds / 60), 60),
      hours   = modf(seconds / 3600),
      days    = days,
      months  = fmod(months, 12),
      years   = modf(months / 12)
      }
  end

  local normalize = interval.normalize
  local timestamptz_add_normalized_interval = timestamptz.add_normalized_interval
  local timestamp_add_normalized_interval = timestamp.add_normalized_interval
  
  function timestamptz.add_interval(ts, int)
    return timestamptz_add_normalized_interval(ts, normalize(int))
  end
  function timestamp.add_interval(ts, int)
    return timestamp_add_normalized_interval(ts, normalize(int))
  end
  function timestamptz.minus_interval(ts, int)
    int = normalize(int)
    int.seconds = - int.seconds
    int.days    = - int.days
    int.months  = - int.months
    return timestamptz_add_normalized_interval(ts, int)
  end
  function timestamp.minus_interval(ts, int)
    int = normalize(int)
    int.seconds = - int.seconds
    int.days    = - int.days
    int.months  = - int.months
    return timestamp_add_normalized_interval(ts, int)
  end
end
do
  local quote = spi.quote
  local to_string = spi.to_string
  local quote_ident = spi.quote_ident
  function spi.to_quoted(val, typ)
    return quote(to_string(val, typ))
  end
  function spi.to_repr(val, typ)
    local typname = typ
    if type(typname) ~= string then
      typname = pg_type.get_name(typname)
    end
    return quote(to_string(val, typ))..'::'..typname
  end
end
do
  local raw = spi.raw
  local qoute = spi.qoute
  function spi.hashstrings_mark_raw(hsh)
    local rawhsh = {}
    for k, v in pairs(hsh) do
      if type(v) == 'string' and v ~= '__mapping__' then
        rawhsh[k] = raw(v)
      else
        rawhsh[k] = v
      end
    end
    return rawhsh
  end
  function spi.hashstrings_quote(hsh)
    local rawhsh = {}
    for k, v in pairs(hsh) do
      if type(v) == 'string' and v ~= '__mapping__' then
        rawhsh[k] = quote(v)
      else
        rawhsh[k] = v
      end
    end
    return rawhsh
  end

  local spi_to_string = spi.to_string
  function spi.eq_values( val1, val2, _type)
    return spi_to_string(val1, _type) == spi_to_string(val2, _type)
  end

  local quote_ident = spi.quote_ident
  function spi.quote_schema_ident(schema, ident)
    if ident == nil then
      local first_dot = find(schema, '%.')
      if first_dot then
        schema, ident = sub(schema, 1, first_dot-1), sub(schema, first_dot+1)
      else
        return quote_ident(schema)
      end
    end
    return format("%s.%s", quote_ident(schema), quote_ident(ident))
  end
end


-- Compile infrastructure
local known_mt = {__metatable = "why do you looks here?"}

-- returns whenever function wants to set field in known table and end-position of such declaration
local function wish_to_set_known(text)
  local start, name_start = find(text, '^%s*%-%- known as ')
  if not start then return nil end
  name_start = name_start + 1
  local name_end, _end = find(text, "%s*\n", name_start+1)
  if not name_end or name_end == name_start  then return nil end
  return sub(text, name_start , name_end - 1), _end + 1
end
local find_known_proc
known = {}

do
  do
    local find_known_proc_query
    local function get_find_known_proc_query()
      local lang_oid = spi.execute([[
          select oid
          from pg_catalog.pg_language 
          where lanname like 'plluac%' and not lanpltrusted]])[1][1]
      return spi.prepare([[
          select oid, prosrc
          from pg_proc
          where prolang = ]] .. lang_oid .. [[ 
            and proisagg = false
            and prosrc like ('%-- known as ' || $1 || '%')
          ]], 'text')
    end
    function find_known_proc(name)
      if not find_known_proc_query then
        find_known_proc_query = get_find_known_proc_query()
      end
      local res = find_known_proc_query:execute(name)
      if res then
        for _, row in ipairs(res) do
          local name_set = wish_to_set_known(row.prosrc)
          if name_set and name_set == name then
            return row.oid
          end
        end
      end
      return nil
    end
  end
  local recursion_check = {}
  local function find_known(knwn, name)
    for i = 1, #recursion_check do
      if recursion_check[i] == name then
        return nil
      end
    end
    local fun_oid = find_known_proc(name)
    if fun_oid then
      compile_pllua_function(fun_oid)
      insert(recursion_check, name)
      local result = knwn[name] -- could be recursion !!!!
      recursion_check[#recursion_check] = nil
      return result
    end
    return nil
  end
  known_mt.__index = find_known
  setmetatable(known, known_mt)
end

local function make_function(str, name, oid, args)
    local all = ''
    local known_name, decl_end = wish_to_set_known(str)
    if known_name then
      local known_oid = find_known_proc(known_name)
      if known_oid and known_oid ~= oid then
        error(format('Dublicate known_name %q (current oid = %q another = %q)', known_name, oid, known_oid))
      end
    end
    local strt, stp = find(str,'^%s*%-%-%[==%[prepare%s-\n', decl_end)
    if strt then
       local lstrt,lstp = find(str,'%-%-%]==%]')
       all = sub(str, stp+1, lstrt-1)
       str = sub(str,lstp+1)
    end
    all = format('%slocal function recursion(%s)\n%s\nend ', all, args, str)
    if known_name then
      all = all .. format(' known[%q] = recursion ', known_name)
    end
    all = all .. 'return recursion'
    return assert(assert(loadstring(all, name))())
end

return make_function
