/***********************************************/
/* this module was written with influence of   */
/* Luis Carvalho code in  project PL/LUA       */
/* http://pgfoundry.org/projects/pllua         */
/* big thanks                                  */
/***********************************************/

#include <pllua.h>

static int luaP_record_index(lua_State * L)
{
    lua_getfield(L, 1, "__mapping__");
    lua_pushvalue(L, 2);
    lua_gettable(L, -2);
    if (lua_isnumber(L, -1)) {
        lua_rawgeti(L, 1, lua_tointeger(L, -1));
    }
    return 1;
}

static int luaP_record_newindex(lua_State * L)
{
    lua_getfield(L, 1, "__mapping__");
    lua_pushvalue(L, 2);
    lua_gettable(L, -2);
    if (lua_isnumber(L, -1)) {
        lua_pushvalue(L, 3);
        lua_rawseti(L, 1, lua_tointeger(L, -2));
    } else {
        lua_settop(L, 3);
        lua_rawset(L, 1);
    }
    return 0;
}

static const luaL_reg luaP_Record_mt[] = {
    {"__index", luaP_record_index},
    {"__newindex", luaP_record_newindex},
    {NULL, NULL}
};

static int luaP_tuptable_index(lua_State * L)
{
    TupleDesc tupdesc = lua_touserdata(L, 2);
    int i;
    const char *key;
//    char * typname;
    int pos = 1;
    lua_createtable(L, 0, tupdesc->natts);
    for (i = 0; i < tupdesc->natts; i++) {
        if (tupdesc->attrs[i]->attisdropped)
            continue;
        key = NameStr(tupdesc->attrs[i]->attname);
        if (strncmp(key, "?column?", 8) != 0) {
            lua_getfield(L, -1, key);
            if (lua_isnil(L, -1)) {
                lua_pushstring(L, key);
            } else {
                lua_pushfstring(L, "%s_%d", key, i);
            }
            lua_replace(L, -2);
        } else
            lua_pushfstring(L, "_%d", i);
        // --first create a {attr_name, typename} table at attr position
        // first create a {attr_name, {type_oid, type_mod}}
        // or {attr_name, type_oid} table at attr position
        lua_createtable(L, 2, 0);
        lua_pushvalue(L, -2);
        lua_rawseti(L, -2, 1);
        if (tupdesc->attrs[i]->atttypmod != -1) {
            lua_createtable(L, 2, 0);
            lua_pushinteger(L, tupdesc->attrs[i]->atttypid);
            lua_rawseti(L, -2, 1);
            lua_pushinteger(L, tupdesc->attrs[i]->atttypmod);
            lua_rawseti(L, -2, 2);
        } else {
            lua_pushinteger(L, tupdesc->attrs[i]->atttypid);
        }
/*
        typname = format_type_with_typemod(
                    tupdesc->attrs[i]->atttypid, 
                    tupdesc->attrs[i]->atttypmod);
        lua_pushstring(L, typname);
        pfree((void*) typname);
*/
        lua_rawseti(L, -2, 2);
        lua_rawseti(L, -3, pos);
        lua_pushinteger(L, pos);
        lua_rawset(L, -3);
        pos++;
    }
    lua_pushvalue(L, 2);
    lua_pushvalue(L, -2);
    lua_settable(L, 1);
    return 1;
}

static const luaL_reg luaP_Tuptable_idx[] = {
    {"__index", luaP_tuptable_index},
    {NULL, NULL}
};
const void *TUPTABLE_META_P = NULL;
const void *TUPTABLE_IDX_P = NULL;

lua_State *LUA_TYPES = NULL;

int luaP_register_conversations(lua_State * L)
{
    LUA_TYPES = lua_newthread(L);
    lua_setfield(L, LUA_REGISTRYINDEX, "PLLUA_TYPES_THREAD");
    lua_pushlightuserdata(L, &TUPTABLE_META_P);
    lua_newtable(L);
    luaL_register(L, NULL, luaP_Record_mt);
    TUPTABLE_META_P = lua_topointer(L, -1);
    lua_pushvalue(L, -1);
    lua_xmove(L, LUA_TYPES, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
    lua_pushlightuserdata(L, &TUPTABLE_IDX_P);
    lua_newtable(L);
    lua_createtable(L, 0, 2);
    luaL_register(L, NULL, luaP_Tuptable_idx);
    lua_pushstring(L, "v");
    lua_setfield(L, -2, "__mode");
    lua_setmetatable(L, -2);
    lua_pushvalue(L, -1);
    lua_xmove(L, LUA_TYPES, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
    return 0;
}

static void luaP_find_tuptable_index(lua_State * L, TupleDesc tupdesc)
{
    lua_pushlightuserdata(LUA_TYPES, tupdesc);
    lua_gettable(LUA_TYPES, 2);
    lua_xmove(LUA_TYPES, L, 1);
}

void luaP_clear_tuptable_index(lua_State * L, TupleDesc tupdesc)
{
    lua_pushlightuserdata(LUA_TYPES, tupdesc);
    lua_pushnil(LUA_TYPES);
    lua_settable(LUA_TYPES, 2);
}


/* VARATT_SIZEP is deprecated in PostgreSQL 8.3 */
#if PG_VERSION_NUM < 80300

#define SET_VARSIZE(ptr, len) VARATT_SIZEP(ptr) = len
#define att_addlength_pointer(len, typ, ptr) att_addlength(len, typ, PointerGetDatum(ptr))
#define att_align_nominal(len, typ) att_align(len, typ)

#endif

static bool luaP_pushdatum_trivial(lua_State * L, Datum dat, Oid type);
static Datum luaP_todatum_trivial(lua_State * L, Oid type, int32 typmod,
                                  bool * isnull, bool * found);


/* Copyright (c) 2008 Luis Carvalho (mostly)*/
static void
luaP_pusharray(lua_State * L, char **p, int ndims,
               int *dims, int *lb, bits8 ** bitmap, int *bitmask,
               Form_pg_type typeinfo, Oid typeelem, cache_type_heap * cache)
{
    int i;
    lua_createtable(L, *dims, 0);
    if (ndims == 1) {           /* vector? */
        for (i = 0; i < (*dims); i++) {
            if (*bitmap == NULL || ((**bitmap) & (*bitmask)) != 0) {    /* not NULL? */
                luaP_pushdatum_cached(L,
                                      fetch_att(*p, typeinfo->typbyval,
                                                typeinfo->typlen), typeelem,
                                      cache);
                lua_rawseti(L, -2, (*lb) + i);
                *p = att_addlength_pointer(*p, typeinfo->typlen, *p);
                *p = (char *) att_align_nominal(*p, typeinfo->typalign);
            } else if (*bitmap != NULL) {
                lua_pushNULL(L);
                lua_rawseti(L, -2, (*lb) + i);
            }
            if (*bitmap) {
                *bitmask <<= 1;
                if (*bitmask == 0x100) {
                    (*bitmap)++;
                    *bitmask = 1;
                }
            }
        }
    } else {                    /* multidimensional array */
        for (i = 0; i < (*dims); i++) {
            luaP_pusharray(L, p, ndims - 1, dims + 1, lb + 1, bitmap, bitmask,
                           typeinfo, typeelem, cache);
            lua_rawseti(L, -2, (*lb) + i);
        }
    }
}

/* Copyright (c) 2008 Luis Carvalho (mostly)*/
void
luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(lua_State * L,
                                                   cache_tuple_desc * cache,
                                                   HeapTuple tup,
                                                   datum_to_lua_mode mode)
{
    int i;
    bool isnull;
    int pos = 1;
    Datum value;
    TupleDesc tupdesc = cache->tupdesc;
    Fpush_datum push_datum;
    switch (mode) {
    case D2L_SIMPLE:
        push_datum = luaP_pushdatum_simple_cached;
        break;
    case D2L_STRINGS:
        push_datum = luaP_pushdatum_string_cached;
        break;
    case D2L_FULL:
    default:
        push_datum = luaP_pushdatum_cached;
        break;
    }

    alloc_cache_tuple_attrs(cache);

    lua_createtable(L, tupdesc->natts, 1);
    for (i = 0; i < tupdesc->natts; i++) {
        if (tupdesc->attrs[i]->attisdropped)
            continue;
        value = heap_getattr(tup, tupdesc->attrs[i]->attnum, tupdesc, &isnull);
        if (!isnull)
            push_datum(L, value, tupdesc->attrs[i]->atttypid,
                       &cache->cache_attr[i]);
        else
            lua_pushNULL(L);
        lua_rawseti(L, -2, pos);
        pos++;
    }
    lua_pushstring(L, "__mapping__");
    luaP_find_tuptable_index(L, tupdesc);
    lua_rawset(L, -3);
    lua_pushlightuserdata(L, &TUPTABLE_META_P);
    lua_rawget(L, LUA_REGISTRYINDEX);
    lua_setmetatable(L, -2);
}

void luaP_pushhash_adv_from_tuple(lua_State * L, TupleDesc tupdesc,
                                  HeapTuple tup, datum_to_lua_mode mode)
{
    CacheTupleDesc(cache);
    cache.tupdesc = tupdesc;
    cache.do_not_release = true;
    luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(L, &cache, tup, mode);
    release_cache_tuple_desc(&cache);
}

void luaP_pushdatum_string_cached(lua_State * L, Datum dat, Oid type,
                                  cache_type_heap * cache)
{
    char *tmp;
    CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
    Form_pg_type typeinfo;
    if (!current->heap) {
        current->heap = SSearchType(type);
    }
    typeinfo = (Form_pg_type) GETSTRUCT(current->heap);
    alloc_cache_type_heap_finfo_out(current);
    tmp = OutputFunctionCall(current->finfo_out, dat);
    lua_pushstring(L, tmp);
    if (current != cache)
        release_cache_type_heap(&localheap);
}


void luaP_pushdatum_cached(lua_State * L, Datum dat, Oid type,
                           cache_type_heap * cache)
{
    bool found = luaP_pushdatum_trivial(L, dat, type);
    if (!found) {
        switch (type) {
        case REFCURSOROID:
            {
                const char *name = text2string(dat);
                Portal cursor = SPI_cursor_find(name);
                pfree((void *) name);
                if (cursor != NULL)
                    luaP_pushcursor(L, cursor);
                else
                    lua_pushnil(L);
                break;
            }
        case INTERVALOID:
            {
                Interval *span = (Interval *) DatumGetIntervalP(dat);
                luaPP_interval_push(L, span);
                break;
            }
        default:
            {
                CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
                Form_pg_type typeinfo;
                if (!current->heap) {
                    current->heap = SSearchType(type);
                }
                typeinfo = (Form_pg_type) GETSTRUCT(current->heap);
                if (typeinfo->typelem != 0 && typeinfo->typlen == -1) {
                    /* we got an array (see catalog.pg_type) */
                    ArrayType *arr = DatumGetArrayTypeP(dat);
                    char *p = ARR_DATA_PTR(arr);
                    bits8 *bitmap = ARR_NULLBITMAP(arr);
                    int bitmask = 1;
                    Oid typeelem = ARR_ELEMTYPE(arr);
                    Form_pg_type typeeleminfo;

                    alloc_cache_type_heap_elem(current);
                    if (!current->cache_elem->heap) {
                        current->cache_elem->heap = SSearchType(typeelem);
                    }
                    typeeleminfo =
                        (Form_pg_type) GETSTRUCT(current->cache_elem->heap);

                    luaP_pusharray(L, &p, ARR_NDIM(arr), ARR_DIMS(arr),
                                   ARR_LBOUND(arr), &bitmap, &bitmask,
                                   typeeleminfo, typeelem, current->cache_elem);
                    if (current != cache) {
                        release_cache_type_heap(current->cache_elem);
                    }
                } else {
                    if (typeinfo->typtype != 'c') {
                        char *tmp;
                        alloc_cache_type_heap_finfo_out(current);
                        tmp = OutputFunctionCall(current->finfo_out, dat);
                        lua_pushstring(L, tmp);
                    } else {
                        HeapTupleData tmptup;
                        alloc_cache_type_heap_tupdesc(current);
                        if (!current->tupdesc.tupdesc) {
                            current->tupdesc.tupdesc =
                                lookup_rowtype_tupdesc(type,
                                                       typeinfo->typtypmod);
                        }

                        tmptup.t_data = DatumGetHeapTupleHeader(dat);
                        tmptup.t_len =
                            HeapTupleHeaderGetDatumLength(tmptup.t_data);
                        ItemPointerSetInvalid(&(tmptup.t_self));
                        tmptup.t_tableOid = InvalidOid;
#if PG_VERSION_NUM >= 80300
                        ItemPointerSetInvalid(&tmptup.t_self);
#endif

                        luaP_pushhash_from_tuple_with_cache_tuple_desc(L,
                                                                       &current->
                                                                       tupdesc,
                                                                       &tmptup);
                        //ReleaseTupleDesc(tupdesc);
                    }
                }
                if (current != cache)
                    release_cache_type_heap(&localheap);
                break;
            }
        }
    }
}

/**** Simple conversions for simplified cases ***/

void luaP_pushdatum_simple_cached(lua_State * L, Datum dat, Oid type,
                                  cache_type_heap * cache)
{
    bool found = luaP_pushdatum_trivial(L, dat, type);
    if (!found) {
        CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
        Form_pg_type typeinfo;
        char *tmp;
        if (!current->heap) {
            current->heap = SSearchType(type);
        }
        typeinfo = (Form_pg_type) GETSTRUCT(current->heap);
        alloc_cache_type_heap_finfo_out(current);
        tmp = OutputFunctionCall(current->finfo_out, dat);
        lua_pushstring(L, tmp);
        if (current != cache)
            release_cache_type_heap(&localheap);
    }
}

/* Trivial cases which doesn't need any type lookups and   */
/* returns only simple values: number, boolean, string     */
static bool luaP_pushdatum_trivial(lua_State * L, Datum dat, Oid type)
{
    switch (type) {
    case OIDOID:
        lua_pushinteger(L, DatumGetObjectId(dat));
        break;
    case BOOLOID:
        lua_pushboolean(L, (int) (dat != 0));
        break;
    case FLOAT4OID:
        lua_pushnumber(L, (lua_Number) DatumGetFloat4(dat));
        break;
    case FLOAT8OID:
        lua_pushnumber(L, (lua_Number) DatumGetFloat8(dat));
        break;
    case INT2OID:
        lua_pushinteger(L, (lua_Integer) DatumGetInt16(dat));
        break;
    case INT4OID:
        lua_pushinteger(L, (lua_Integer) DatumGetInt32(dat));
        break;
    case INT8OID:
        lua_pushnumber(L, (lua_Integer) DatumGetInt64(dat));
        break;
    case NUMERICOID:
        {
            Datum fl;
            fl = DirectFunctionCall1(numeric_float8, dat);
            lua_pushnumber(L, DatumGetFloat8(fl));
        }
        break;
    case TEXTOID:{
            const char *str = text2string(dat);
            lua_pushstring(L, str);
            pfree((void *) str);
        }
        break;
    case BPCHAROID:{
            const char *str = char2string(dat);
            lua_pushstring(L, str);
            pfree((void *) str);
        }
        break;
    case VARCHAROID:{
            const char *str = varchar2string(dat);
            lua_pushstring(L, str);
            pfree((void *) str);
        }
        break;
    case TIMESTAMPOID:
    case TIMESTAMPTZOID:
        {
            TimestampTz ts = (TimestampTz) DatumGetTimestampTz(dat);
            luaP_pushtimestamp(L, ts);
            break;
        }
    default:
        {
            return false;
            break;
        }
    }
    return true;
}


/**********************************/
/*                                */
/* ======= luaP_getresult ======= */
/*                                */
/**********************************/

/* assume table is in top, returns size */
static void
luaP_getarraydims(lua_State * L, int *ndims, int *dims,
                  int *lb, bool * hasnulls)
{
    int nitems = 0;
    *ndims = -1;
    *hasnulls = false;
    lua_pushnil(L);
    while (lua_next(L, -2)) {
        if (lua_isnumber(L, -2)) {
            int n = 0;
            int k = lua_tointeger(L, -2);
            /* set dims and lb */
            if (*ndims < 0) {   /* first visit? */
                *lb = k;
                *dims = 1;
            } else {
                if (*lb > k) {
                    *dims += *lb - k;
                    *lb = k;
                }
                if (*dims - 1 + *lb < k)        /* max < k? */
                    *dims = k - *lb + 1;
            }
            /* set ndims */
            if (lua_type(L, -1) == LUA_TTABLE) {
                int not_isrecord;
                lua_getfield(L, -1, "__mapping__");     /* to allow numeric tables acts as record */
                not_isrecord = lua_isnil(L, -1);
                lua_pop(L, 1);
                if (not_isrecord) {
                    int d = -1, l = -1;
                    if (*ndims == MAXDIM)
                        elog(ERROR,
                             "[pllua]: table exceeds max number of dimensions");
                    if (*ndims > 1) {
                        d = dims[1];
                        l = lb[1];
                    }
                    luaP_getarraydims(L, &n, dims + 1, lb + 1, hasnulls);
                    if (n >= 0) {
                        if (*ndims >= 1) {      /* update dims and bounds? */
                            if (l < lb[1]) {
                                lb[1] = l;
                                *hasnulls = true;
                            }
                            if (d + l > dims[1] + lb[1]) {
                                dims[1] = d + l - lb[1];
                                *hasnulls = true;
                            }
                        }
                    } else {
                        n = 0;
                    }
                }
            }
            n++;
            if (*ndims < 0)
                *ndims = n;
            else if (*ndims != n)
                elog(ERROR, "[pllua]: table is asymetric");
        }
        nitems++;
        lua_pop(L, 1);
    }
    if (!(*hasnulls))
        *hasnulls = (nitems > 0 && nitems != *dims);
}

static void
luaP_collect_datums_toarray(lua_State * L,
                            int ndims, int *dims, int *lb,
                            Datum ** vals, bool ** nulls,
                            Oid typeelem, int typmod, cache_type_heap * cache)
{
    int i;
    if (ndims == 1) {
        for (i = 0; i < (*dims); i++, (*vals)++, (*nulls)++) {
            lua_rawgeti(L, -1, (*lb) + i);
            **vals = luaP_todatum_cached(L, typeelem, typmod, *nulls, cache);
            lua_pop(L, 1);
        }
    } else {
        for (i = 0; i < (*dims); i++) {
            lua_rawgeti(L, -1, (*lb) + i);
            luaP_collect_datums_toarray(L, ndims - 1, dims + 1, lb + 1,
                                        vals, nulls, typeelem, typmod, cache);
            lua_pop(L, 1);
        }
    }
}

HeapTuple luaP_form_tuple_with_tuple_desc(lua_State * L, TupleDesc tupdesc)
{
    HeapTuple tup = NULL;
    CacheTupleDesc(cache);
    cache.tupdesc = tupdesc;
    cache.do_not_release = true;
    tup = luaP_form_tuple_with_cache_tuple_desc(L, &cache);
    release_cache_tuple_desc(&cache);
    return tup;
}

HeapTuple
luaP_form_tuple_with_cache_tuple_desc(lua_State * L, cache_tuple_desc * cache)
{
    int i;
    Datum *values;
    bool *nulls;
    HeapTuple tup;
    TupleDesc tupdesc = cache->tupdesc;

    if (lua_type(L, -1) != LUA_TTABLE)
        elog(ERROR, "[pllua]: table expected for record result, got %s",
             lua_typename(L, lua_type(L, -1)));
    /* create tuple */
    alloc_cache_tuple_attrs(cache);
    values = palloc0(tupdesc->natts * sizeof(Datum));
    nulls = palloc0(tupdesc->natts * sizeof(bool));
    for (i = 0; i < tupdesc->natts; i++) {
        if (tupdesc->attrs[i]->attisdropped) {
            nulls[i] = true;
            values[i] = PointerGetDatum(NULL);;
            continue;
        }
        lua_getfield(L, -1, NameStr(tupdesc->attrs[i]->attname));
        values[i] = luaP_todatum_cached(L, tupdesc->attrs[i]->atttypid,
                                        tupdesc->attrs[i]->atttypmod,
                                        nulls + i, &cache->cache_attr[i]);
        lua_pop(L, 1);
    }
    tup = heap_form_tuple(tupdesc, values, nulls);
    pfree(nulls);
    pfree(values);
    return tup;
}

Datum luaP_string_todatum_cached(lua_State * L, Oid type, int32 typmod,
                                 bool * isnull, cache_type_heap * cache)
{
    Datum dat = 0;
    const char *tmp = NULL;
    *isnull = lua_isNULL(L, -1);
    if (!(*isnull || type == VOIDOID)) {
        CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
        tmp = lua_tostring(L, -1);
        if (tmp == NULL) {
            *isnull = true;
        } else {
            if (!current->heap) {
                current->heap = SSearchType(type);
            }
            alloc_cache_type_heap_finfo_in(current);
            dat =
                InputFunctionCall(current->finfo_in, (char *) tmp,
                                  getTypeIOParam(current->heap), typmod);
        }
        if (current != cache) {
            release_cache_type_heap(&localheap);
        }
    }
    return dat;
}

#define lua_IntegerMin  PTRDIFF_MIN
#define lua_IntegerMax  PTRDIFF_MAX

Datum
luaP_todatum_cached(lua_State * L, Oid type, int32 typmod, bool * isnull,
                    cache_type_heap * cache)
{
    Datum dat = 0;              /* NULL */
    bool found = false;
    if (luaP_israwstring(L, -1)) {
        lua_rawgeti(L, -1, 1);
        dat = luaP_string_todatum_cached(L, type, typmod, isnull, cache);
        lua_pop(L, 1);
        return dat;
    }
    dat = luaP_todatum_trivial(L, type, typmod, isnull, &found);
    if (!found) {
        switch (type) {
        case REFCURSOROID:
            {
                luaP_cursor *cursor = luaP_tocursor(L, -1);
                dat = string2text(cursor->p->name);
                break;
            }
        case INTERVALOID:
            {
                Interval *span = (Interval *) palloc0(sizeof(Interval));
                luaPP_interval_normalize(L, -1, span);
                dat = IntervalPGetDatum(span);
                break;
            }
        default:
            {
                CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
                Form_pg_type typeinfo;
                if (!current->heap) {
                    current->heap = SSearchType(type);
                }
                typeinfo = (Form_pg_type) GETSTRUCT(current->heap);

                if (typeinfo->typelem != 0 && typeinfo->typlen == -1) { /* got an array */
                    Oid typeelem = typeinfo->typelem;
                    int ndims, dims[MAXDIM], lb[MAXDIM];
                    int i, nitems;
                    bool hasnulls;
                    ArrayType *ar;
                    if (lua_type(L, -1) != LUA_TTABLE)
                        elog(ERROR,
                             "[pllua]: table expected for array conversion, got %s",
                             lua_typename(L, lua_type(L, -1)));
                    for (i = 0; i < MAXDIM; i++)
                        dims[i] = lb[i] = -1;
                    luaP_getarraydims(L, &ndims, dims, lb, &hasnulls);
                    nitems = ArrayGetNItems(ndims, dims);
                    if (nitems == 0) {  /* empty array? */
                        ar = construct_empty_array(typeelem);
                    } else {
                        Datum *vals = (Datum *) palloc(sizeof(Datum) * nitems);
                        bool *nulls = (bool *) palloc(sizeof(bool) * nitems);
                        Datum *tvals = vals;
                        bool *tnulls = nulls;
                        Form_pg_type typeeleminfo;

                        alloc_cache_type_heap_elem(current);
                        if (!current->cache_elem->heap) {
                            current->cache_elem->heap = SSearchType(typeelem);
                        }
                        typeeleminfo =
                            (Form_pg_type) GETSTRUCT(current->cache_elem->heap);

                        luaP_collect_datums_toarray(L, ndims, dims, lb,
                                                    &tvals, &tnulls,
                                                    typeelem, typmod,
                                                    current->cache_elem);
                        ar = construct_md_array(vals, nulls, ndims, dims, lb,
                                                typeelem, typeeleminfo->typlen,
                                                typeeleminfo->typbyval,
                                                typeeleminfo->typalign);
                        pfree(vals);
                        pfree(nulls);
                        if (current != cache) {
                            release_cache_type_heap(current->cache_elem);
                        }
                    }
                    dat = PointerGetDatum(ar);
                } else {
                    if (typeinfo->typtype != 'c') {     // not a record
                        char *tmp;
                        lua_pushvalue(L, -1);
                        tmp = (char *) lua_tostring(L, -1);
                        if (tmp == NULL) {
                            lua_pop(L, 1);
                            *isnull = true;
                            if (current != cache)
                                release_cache_type_heap(&localheap);
                            break;
                        }
                        alloc_cache_type_heap_finfo_in(current);
                        dat =
                            InputFunctionCall(current->finfo_in, tmp,
                                              getTypeIOParam(current->heap),
                                              typmod);
                        lua_pop(L, 1);
                    } else {
                        HeapTuple tuple = NULL;
                        HeapTupleHeader tupdat = NULL;

                        alloc_cache_type_heap_tupdesc(current);
                        if (!current->tupdesc.tupdesc) {
                            current->tupdesc.tupdesc =
                                lookup_rowtype_tupdesc(type,
                                                       typeinfo->typtypmod);
                        }

                        tuple =
                            luaP_form_tuple_with_cache_tuple_desc(L,
                                                                  &current->
                                                                  tupdesc);
                        tupdat = (HeapTupleHeader) palloc(tuple->t_len);
                        memcpy(tupdat, tuple->t_data, tuple->t_len);

                        dat = PointerGetDatum(tupdat);
                        heap_freetuple(tuple);
                    }
                }
                if (current != cache)
                    release_cache_type_heap(&localheap);
            }
        }
    }
    return dat;
}


/*** Simple conversions for simplified cases **/

Datum
luaP_todatum_simple_cached(lua_State * L, Oid type, int32 typmod, bool * isnull,
                           cache_type_heap * cache)
{
    Datum dat = 0;              /* NULL */
    bool found = false;
    if (luaP_israwstring(L, -1)) {
        lua_rawgeti(L, -1, 1);
        dat = luaP_string_todatum_cached(L, type, typmod, isnull, cache);
        lua_pop(L, 1);
        return dat;
    }
    dat = luaP_todatum_trivial(L, type, typmod, isnull, &found);
    if (!found) {
        char *tmp;
        lua_pushvalue(L, -1);
        tmp = (char *) lua_tostring(L, -1);
        if (tmp == NULL) {
            *isnull = true;
        } else {
            CacheTypeHeap(localheap), *current = cache ? cache : &localheap;
            Form_pg_type typeinfo;
            alloc_cache_type_heap_finfo_in(current);
            if (!current->heap) {
                current->heap = SSearchType(type);
            }
            typeinfo = (Form_pg_type) GETSTRUCT(current->heap);
            dat =
                InputFunctionCall(current->finfo_in, tmp,
                                  getTypeIOParam(current->heap), typmod);
            if (current != cache)
                release_cache_type_heap(&localheap);
        }
        lua_pop(L, 1);
    }
    return dat;
}


/*  Trivial cases which doesn't need any type lookups 
    and gots only simple lua values: number, boolean, string */
static Datum luaP_todatum_trivial(lua_State * L, Oid type, int32 typmod,
                                  bool * isnull, bool * found)
{
    Datum dat = 0;              /* NULL */
    *isnull = lua_isNULL(L, -1);
    if (!(*isnull || type == VOIDOID)) {
        switch (type) {
        case OIDOID:
            dat = ObjectIdGetDatum(lua_tointeger(L, -1));
            break;
        case BOOLOID:
            dat = BoolGetDatum(lua_toboolean(L, -1));
            break;
        case FLOAT4OID:
            dat = Float4GetDatum(lua_tonumber(L, -1));
            break;
        case FLOAT8OID:
            dat = Float8GetDatum(lua_tonumber(L, -1));
            break;
        case INT2OID:
            dat = Int16GetDatum(lua_tointeger(L, -1));
            break;
        case INT4OID:
            dat = Int32GetDatum(lua_tointeger(L, -1));
            break;
        case INT8OID:
            {                   /* not by value */
                lua_Number n = lua_tonumber(L, -1);
                int64 i;
                if (n >= lua_IntegerMin && n <= lua_IntegerMax) {
                    i = lua_tointeger(L, -1);
                } else if (n <= INT64_MAX && n >= INT64_MIN) {
                    i = n;
                } else if (n < INT64_MIN) {
                    i = INT64_MIN;
                } else {
                    i = INT64_MAX;
                }
                dat = Int64GetDatum(i);
                break;
            }
        case NUMERICOID:
            dat =
                DirectFunctionCall1(float8_numeric,
                                    Float8GetDatum(lua_tonumber(L, -1)));
            break;
        case TEXTOID:
            {
                const char *ch = lua_tostring(L, -1);
                if (ch != NULL)
                    dat = string2text(ch);
                else
                    *isnull = true;
            }
            break;
        case BPCHAROID:{
                const char *ch = lua_tostring(L, -1);
                if (typmod <= 0)
                    elog(ERROR,
                         "[pllua]: type '%s' not supported in this context",
                         format_type_be(type));
                if (ch != NULL)
                    dat = string2char(ch, typmod);
                else
                    *isnull = true;
            }
            break;
        case VARCHAROID:{
                const char *ch = lua_tostring(L, -1);
                if (ch != NULL)
                    dat = string2varchar(ch, typmod);
                else
                    *isnull = true;
            }
            break;
        case TIMESTAMPOID:
        case TIMESTAMPTZOID:
            {
                TimestampTz ts = luaP_totimestamp(L, -1);
                dat = TimestampTzGetDatum(ts);
                break;
            }
        default:
            {
                *found = false;
                return 0;
                break;
            }
        }
    }
    *found = true;
    return dat;
}

/*************************************************
 * Working with cache
 *************************************************/

#define InitFuncContext() \
    MemoryContext oldcontext = current_call_info ? MemoryContextSwitchTo(current_call_info->func_context) : NULL; \
/*       ResourceOwner oldowner = CurrentResourceOwner; */ \
                                /*       CurrentResourceOwner = current_call_info->func_resource */

#define ReleaseFuncContext() \
    if (current_call_info) MemoryContextSwitchTo(oldcontext); \
                                /*       CurrentResourceOwner = oldowner */

void release_cache_tuple_desc(cache_tuple_desc * cache)
{
    int i;
    if (cache->tupdesc) {
        InitFuncContext();
        if (cache->cache_attr) {
            for (i = 0; i < cache->tupdesc->natts; i++) {
                if (cache->cache_attr[i].heap)
                    release_cache_type_heap(&cache->cache_attr[i]);
            }
            pfree(cache->cache_attr);
        }
        if (!cache->do_not_release) {
            ReleaseTupleDesc(cache->tupdesc);
            MemSet(cache, 0, sizeof(cache_tuple_desc));
        } else {
            cache->cache_attr = NULL;
        }
        ReleaseFuncContext();
    }
};

void alloc_cache_tuple_attrs(cache_tuple_desc * cache)
{
    if (cache->tupdesc && !cache->cache_attr) {
        InitFuncContext();
        cache->cache_attr =
            (cache_type_heap *) palloc0(sizeof(cache_type_heap) *
                                        cache->tupdesc->natts);;
        ReleaseFuncContext();
    }
};

void alloc_cache_type_heap_elem(cache_type_heap * cache)
{
    if (!cache->cache_elem) {
        InitFuncContext();
        cache->cache_elem =
            (cache_type_heap *) palloc0(sizeof(cache_type_heap));
        ReleaseFuncContext();
    }
}

void alloc_cache_type_heap_tupdesc(cache_type_heap * cache)
{

}

void alloc_cache_type_heap_finfo_in(cache_type_heap * cache)
{
    Form_pg_type typeinfo = (Form_pg_type) GETSTRUCT(cache->heap);
    if (!cache->finfo_in) {
        InitFuncContext();
        cache->finfo_in = (FmgrInfo *) palloc0(sizeof(FmgrInfo));
        fmgr_info(typeinfo->typinput, cache->finfo_in);
        ReleaseFuncContext();
    }
}

void alloc_cache_type_heap_finfo_out(cache_type_heap * cache)
{
    Form_pg_type typeinfo = (Form_pg_type) GETSTRUCT(cache->heap);
    if (!cache->finfo_out) {
        InitFuncContext();
        cache->finfo_out = (FmgrInfo *) palloc0(sizeof(FmgrInfo));
        fmgr_info(typeinfo->typoutput, cache->finfo_out);
        ReleaseFuncContext();
    }
}

void release_cache_type_heap(cache_type_heap * cache)
{
    InitFuncContext();
    if (cache->tupdesc.tupdesc) {
        release_cache_tuple_desc(&cache->tupdesc);
    }
    if (cache->cache_elem) {
        release_cache_type_heap(cache->cache_elem);
        pfree(cache->cache_elem);
    }
    if (cache->finfo_in) {
        pfree(cache->finfo_in);
    }
    if (cache->finfo_out) {
        pfree(cache->finfo_out);
    }
    if (cache->heap) {
        ReleaseSysCache(cache->heap);
    }
    MemSet(cache, 0, sizeof(cache_type_heap));
    ReleaseFuncContext();
}
