#include <pllua.h>

static void init_tuple_store_cachetupdesc();
static void push_tuple_to_store(lua_State * LUA);

/*
 * Check two tupledescs have matching number and types of attributes
 */
#define VARCHARARRAYOID 1015
#define TEXTARRAYOID 1009
static bool compatible_tupdesc(TupleDesc td1, TupleDesc td2)
{
    int i;

    if (td1->natts != td2->natts)
        return false;

    for (i = 0; i < td1->natts; i++) {
        if (td1->attrs[i]->atttypid != td2->attrs[i]->atttypid) {
            Oid oid1 = td1->attrs[i]->atttypid;
            Oid oid2 = td2->attrs[i]->atttypid;
            if (((oid1 == VARCHAROID || oid1 == TEXTOID)
                 && (oid2 == VARCHAROID || oid2 == TEXTOID))
                || ((oid1 == VARCHARARRAYOID || oid1 == TEXTARRAYOID)
                    && (oid2 == VARCHARARRAYOID || oid2 == TEXTARRAYOID)))
                continue;
            else
                return false;
        }
    }

    return true;
}


static void init_tuple_store_cachetupdesc()
{
    ReturnSetInfo *rsi =
        (ReturnSetInfo *) current_call_info->fcinfo->resultinfo;
    if (!current_call_info->tuple_store) {
        MemoryContext cxt =
            MemoryContextSwitchTo(rsi->econtext->ecxt_per_query_memory);
        current_call_info->tuple_store =
            tuplestore_begin_heap(true, false, work_mem);
        if (!current_call_info->tuple_store)
            elog(ERROR, "CANT GET current tuple store");
        current_call_info->cache_heap.tupdesc.tupdesc =
            CreateTupleDescCopy(rsi->expectedDesc);
        current_call_info->cache_heap.tupdesc.do_not_release = true;
        MemoryContextSwitchTo(cxt);
    }
}

static HeapTuple heap_header_form_tuple(TupleDesc tupleDescriptor,
                                        HeapTupleHeader td)
{
    HeapTuple tuple;            /* return tuple */
    Size len;

    len = HeapTupleHeaderGetDatumLength(td);
    tuple = (HeapTuple) palloc0(HEAPTUPLESIZE);
    tuple->t_data = td;
    tuple->t_len = len;
    ItemPointerSetInvalid(&(tuple->t_self));
    tuple->t_tableOid = InvalidOid;

    return tuple;
}

static void push_tuple_to_store(lua_State * LUA)
{
    HeapTuple tuple;
    HeapTupleHeader tuphead = NULL;
    ReturnSetInfo *rsi =
        (ReturnSetInfo *) current_call_info->fcinfo->resultinfo;
    if (current_call_info->fn_retistuple) {
        if (luaP_israwstring(LUA, -1)) {
            bool isnull;
            TupleDesc tupdesc = current_call_info->cache_heap.tupdesc.tupdesc;
            Datum dat = luaP_todatum_cached(LUA,
                                            tupdesc->tdtypeid,
                                            tupdesc->tdtypmod,
                                            &isnull,
                                            &current_call_info->cache_heap);

            tuphead = DatumGetHeapTupleHeader(dat);
            tuple = heap_header_form_tuple(tupdesc, tuphead);
        } else if (lua_istable(LUA, -1)) {
            tuple =
                luaP_form_tuple_with_cache_tuple_desc(LUA,
                                                      &current_call_info->
                                                      cache_heap.tupdesc);
        } else {
            ereport(ERROR,
                    (errcode(ERRCODE_DATATYPE_MISMATCH),
                     errmsg
                     ("setof-composite-returning Lua function "
                      "must returns or yields a table, got %s",
                      luaL_typename(LUA, -1))));
            return;
        }
    } else {
        Datum ret;
        bool isNull;
        if (lua_istable(LUA, -1)) {
            lua_getfield(LUA, -1, "__mapping__");
            if (!lua_isnil(LUA, -1)) {
                lua_pop(LUA, 1);
                lua_rawgeti(LUA, -1, 1);
                lua_replace(LUA, -2);
            } else {
                lua_pop(LUA, 1);
            }
        }
        if (lua_isNULL(LUA, -1)) {
            ret = (Datum) 0;
            isNull = true;
        } else {
            ret =
                luaP_todatum_cached(LUA, current_call_info->result_oid, -1,
                                    &isNull,
                                    current_call_info->cache_heap.tupdesc.
                                    cache_attr);
        }

        tuple =
            heap_form_tuple(current_call_info->cache_heap.tupdesc.tupdesc, &ret,
                            &isNull);
    }
    {
        MemoryContext cxt =
            MemoryContextSwitchTo(rsi->econtext->ecxt_per_query_memory);
        tuplestore_puttuple(current_call_info->tuple_store, tuple);
        MemoryContextSwitchTo(cxt);
    }
    heap_freetuple(tuple);
    lua_pop(LUA, 1);
}

int luaP_return_next(lua_State * L)
{
    ReturnSetInfo *rsi =
        (ReturnSetInfo *) current_call_info->fcinfo->resultinfo;
    if (!current_call_info->prodesc->fn_retisset) {
        lua_pushstring(L, "return next called in a non set context");
        lua_error(L);
    }
    if (lua_gettop(L) == 0 || lua_isnil(L, 1))
        return 0;
    if (luaP_iscursor(L, 1)) {
        luaP_cursor *cursor = luaP_tocursor(LUA, 1);
        bool compatible = false;
        SPI_cursor_fetch(cursor->p, true, 1);
        if (SPI_processed) {
            int i;
            init_tuple_store_cachetupdesc();

            compatible =
                compatible_tupdesc(SPI_tuptable->tupdesc,
                                   current_call_info->cache_heap.tupdesc.
                                   tupdesc);
            if (compatible) {
                while (SPI_processed) {
                    MemoryContext cxt =
                        MemoryContextSwitchTo(rsi->econtext->
                                              ecxt_per_query_memory);
                    /* elog(NOTICE, "Running fast copy"); */
                    for (i = 0; i < SPI_processed; i++) {
                        HeapTuple tuple = heap_copytuple(SPI_tuptable->vals[i]);
                        tuplestore_puttuple(current_call_info->tuple_store,
                                            tuple);
                    }
                    MemoryContextSwitchTo(cxt);
                    SPI_freetuptable(SPI_tuptable);
                    SPI_cursor_fetch(cursor->p, true, 100);     //FETCH_ALL);
                }
            } else {
                CacheTupleDesc(cache);
                cache.tupdesc = SPI_tuptable->tupdesc;
                cache.do_not_release = true;
                while (SPI_processed) {
                    for (i = 0; i < SPI_processed; i++) {
                        luaP_pushhash_from_tuple_with_cache_tuple_desc(L,
                                                                       &cache,
                                                                       SPI_tuptable->
                                                                       vals[i]);
                        push_tuple_to_store(LUA);
                    }
                    SPI_freetuptable(SPI_tuptable);
                    SPI_cursor_fetch(cursor->p, true, 100);     //FETCH_ALL);
                }
                release_cache_tuple_desc(&cache);
            }
        }
        SPI_cursor_close(cursor->p);
        cursor->closed = true;
        return 0;
    }
    if (lua_istable(L, 1)) {
        if (luaP_isresultset(L, 1)) {
            int i = 1;
            lua_getfield(L, LUA_REGISTRYINDEX, "RETURN_NEXT");
            do {
                lua_pushvalue(L, -1);
                lua_rawgeti(L, 1, i);
                if (lua_isnil(L, -1))
                    return 0;
                lua_call(L, 1, 0);
                i++;
            } while (true);
        }
        lua_getfield(L, 1, "iterate");
        if (lua_isfunction(L, -1)) {
            lua_pushvalue(L, -1);
            lua_pushvalue(L, 1);
            lua_replace(L, 2);
            lua_replace(L, 1);
            lua_settop(L, 2);
            lua_call(L, 1, LUA_MULTRET);
        } else {
            lua_pop(L, 1);
        }
    }
    if (lua_gettop(L) > 1 && lua_gettop(L) <= 3) {
        if (!lua_isfunction(L, 1)) {
            ereport(ERROR,
                    (errcode(ERRCODE_DATATYPE_MISMATCH),
                     errmsg
                     ("set-returning Lua function must return "
                      "iterator function")));
        }
    }
    if (lua_isfunction(L, 1)) { /* got an iteration function */
        do {
            lua_settop(L, 3);   /* other values ignored */
            lua_pushvalue(LUA, 1);      /* function */
            lua_pushvalue(LUA, 2);      /* object */
            lua_pushvalue(LUA, 3);      /* index */
            /* call iterator function */
            /* function at cell 4, object - 5, index - 6 */
            lua_call(LUA, 2, 2);
            /* expects index at cell 4 and result at cell 5 */
            lua_settop(LUA, 5);
            lua_pushvalue(LUA, 4);
            lua_replace(LUA, 3);        /* set index to watched */
            if (lua_isnil(L, 3))
                break;
            lua_getfield(L, LUA_REGISTRYINDEX, "RETURN_NEXT");  /* prepare to recurse */
            lua_replace(LUA, 4);
            lua_call(LUA, 1, 0);        /* recurse */
        } while (true);
    } else {
        if (current_call_info->fn_retistuple &&
            !lua_istable(LUA, 1) && !luaP_israwstring(LUA, 1)) {
            ereport(ERROR,
                    (errcode(ERRCODE_DATATYPE_MISMATCH),
                     errmsg
                     ("setof-composite-returning Lua function "
                      "must returns or yields a table or raw, got %s",
                      luaL_typename(LUA, -1))));
        }
        init_tuple_store_cachetupdesc();
        push_tuple_to_store(L);
    }
    return 0;
}
