set names 'sql_ascii';
set lc_messages='C';

create or replace function plluacu_ts_1(ts timestamp)
returns timestamp
language plluacu
as $$
  return ts
$$;

create or replace function plluacu_tsz_1(ts timestamptz)
returns timestamptz
language plluacu
as $$
  return ts
$$;

select 'plluacu_ts_1('''||t||''')', plluacu_ts_1(t)
from (values 
  (now()::timestamp),
  ('now'::timestamp),
  ('infinity'::timestamp),
  ('-infinity'::timestamp)) as tt(t);

\echo testing microseconds - year 2240 still works

select 'plluacu_tsz_1('''||(tnow + t*'1 microsecond'::interval)||''')', plluacu_tsz_1(tnow + t*'1 microsecond'::interval)
from generate_series(1, 100000) as t,
    (values ('2008-01-01 01:01:01.010101'::timestamptz),
            ('2240-01-01'::timestamptz)
      ) as tt(tnow)
where tnow + t*'1 microsecond'::interval <> plluacu_tsz_1(tnow + t*'1 microsecond'::interval);


\echo testing 10 microseconds - year 2440 still works

select 'plluacu_tsz_1('''||(tnow + t*'10 microsecond'::interval)||''')', plluacu_tsz_1(tnow + t*'10 microsecond'::interval)
from generate_series(1, 100000) as t,
    (values ('2008-01-01 01:01:01.010101'::timestamptz),
            ('2440-01-01'::timestamptz)
      ) as tt(tnow)
where tnow + t*'10 microsecond'::interval <> plluacu_tsz_1(tnow + t*'10 microsecond'::interval);


\echo testing milliseconds - year 3000 still works

select 'plluacu_tsz_1('''||(tnow + t*'1 millisecond'::interval)||''')',
        plluacu_tsz_1(tnow + t*'1 millisecond'::interval),
        plluacu_tsz_1(plluacu_tsz_1(tnow + t*'1 millisecond'::interval))
from generate_series(1, 100000) as t,
    (values ('2008-01-01 01:01:01.010101'::timestamptz),
            ('3000-01-01'::timestamptz)
      ) as tt(tnow)
where tnow + t*'1 millisecond'::interval <> plluacu_tsz_1(tnow + t*'1 millisecond'::interval);


\echo timezone difference

create or replace function plluacu_ts_tsz_1(ts timestamp)
returns timestamptz
language plluacu
as $$
  return ts
$$;

create or replace function plluacu_tsz_ts_1(ts timestamptz)
returns timestamp
language plluacu
as $$
  return ts
$$;

select  tnow::timestamp,
        plluacu_ts_tsz_1(tnow::timestamp),
        tnow,
        plluacu_tsz_ts_1(tnow)
from (values ('2008-01-01 01:01:01.010101'::timestamptz), ('3000-01-01'), ('1900-01-01'), ('1800-01-01')  ) as tt(tnow);

create or replace function plluacu_ts_tsz_2(ts timestamp)
returns timestamptz
language plluacu
as $$
  ts = timestamp.to_timestamptz(ts)
  return ts
$$;

create or replace function plluacu_tsz_ts_2(ts timestamptz)
returns timestamp
language plluacu
as $$
  ts = timestamptz.to_timestamp(ts)
  return ts
$$;

select  tnow::timestamp,
        plluacu_ts_tsz_2(tnow::timestamp),
        tnow,
        plluacu_tsz_ts_2(tnow)
from (values ('2008-01-01 01:01:01.010101'::timestamptz), ('3000-01-01'), ('1900-01-01'), ('1800-01-01') ) as tt(tnow);


create or replace function plluacu_tsz_minus(timestamptz, timestamptz)
returns float8
language plluacu
as $$
  spi.info(_1)
  spi.info(_2)
  spi.info(spi.to_string(_1, 'float8'))
  spi.info(spi.to_string(_2, 'float8'))
  spi.info(spi.to_string(_1, 'timestamptz'))
  spi.info(spi.to_string(_2, 'timestamp'))
  return _1 - _2
$$;
select extract(epoch from '2008-05-01 00:00:00+04'::timestamptz - '2008-01-01 00:00:00+03'::timestamptz),
       plluacu_tsz_minus('2008-05-01 00:00:00+04'::timestamptz, '2008-01-01 00:00:00+03'::timestamptz);


create or replace function plluacu_tsz_get_epoch(timestamptz)
returns float8
language plluacu
as $$
  local diff= _1 - timestamptz.epoch
  spi.info(_1)
  spi.info(diff)
  return diff
$$;
select extract(epoch from '2008-05-01 00:00:00+04'::timestamptz),
       plluacu_tsz_get_epoch('2008-05-01 00:00:00+04'::timestamptz);


create or replace function plluacu_ts_3()
returns timestamp
language plluacu
as $$
  return spi.parse('2008-05-01 00:00:00.111111', 'timestamp')
$$;
select plluacu_ts_3();

create or replace function plluacu_tsz_now()
returns timestamptz
language plluacu
as $$
  return timestamptz.now()
$$;

select plluacu_tsz_now(), now();

create or replace function plluacu_ts_to_char()
returns void
language plluacu
as $$
  local info = spi.info
  local function tszinfo(ts)
    info(timestamptz.to_char(ts, "HH24:MI:SS.US Day DD Mon YYYY TZ"))
  end
  local function tsinfo(ts)
    info(timestamp.to_char(ts, "HH24:MI:SS.US Day DD Mon YYYY TZ"))
  end
  local nowz = timestamptz.now()
  local now = timestamp.now()
  tszinfo(nowz)
  tsinfo(now)
  local good, val = pcall(timestamptz.trunc, nowz, 'gom')
  if good then
    tszinfo(val)
  else
    info(val)
  end
  tszinfo(timestamptz.trunc(nowz, 'minute'))
  tszinfo(timestamptz.trunc(nowz, 'hour'))
  tszinfo(timestamptz.trunc(nowz, 'day'))
  tszinfo(timestamptz.trunc(nowz, 'month'))
  tszinfo(timestamptz.trunc(nowz, 'year'))
  tsinfo(timestamp.trunc(now, 'minute'))
  tsinfo(timestamp.trunc(now, 'hour'))
  tsinfo(timestamp.trunc(now, 'day'))
  tsinfo(timestamp.trunc(now, 'month'))
  tsinfo(timestamp.trunc(now, 'year'))
  info(timestamptz.part(nowz, 'second'))
  info(timestamptz.part(nowz, 'minute'))
  info(timestamptz.part(nowz, 'hour'))
  info(timestamptz.part(nowz, 'day'))
  info(timestamptz.part(nowz, 'month'))
  info(timestamptz.part(nowz, 'year'))
  info(timestamp.part(now, 'second'))
  info(timestamp.part(now, 'minute'))
  info(timestamp.part(now, 'hour'))
  info(timestamp.part(now, 'day'))
  info(timestamp.part(now, 'month'))
  info(timestamp.part(now, 'year'))
  local tmz = timestamptz.to_tm(nowz)
  for k, v in pairs(tmz) do
    info("tm." .. k .. " = " .. v )
  end
  tszinfo(nowz)
  tszinfo(timestamptz.from_tm(tmz))
  tmz.tz = nil;
  tszinfo(timestamptz.from_tm(tmz))
  tsinfo(timestamptz.from_tm(tmz))
$$;

select plluacu_ts_to_char();

create or replace function plluacu_ts_pl_interval(a timestamp, b interval)
returns timestamp
language plluacu
as $$
  return timestamp.add_interval(a, b)
$$;

create or replace function plluacu_tsz_pl_interval(a timestamptz, b interval)
returns timestamptz
language plluacu
as $$
  spi.info(spi.to_repr(b, 'interval') .. ' = {seconds = ' .. b.seconds .. 
           ', days = '    .. b.days ..
           ', months = '  .. b.months .. '}')
  return timestamptz.add_interval(a, b)
$$;

select '2008-01-01'::timestamp,   plluacu_ts_pl_interval('2008-01-01', '1 month'::interval);
select '2008-01-01'::timestamptz, 
        plluacu_tsz_pl_interval('2008-01-01'::timestamptz, interv)
from (values 
        ('1 millisecond'::interval),
        ('1 second'::interval),
        ('1 minute'::interval),
        ('1 hour'::interval),
        ('1 day'::interval),
        ('1 month'::interval),
        ('1 year'::interval),
        ('1 year 4 months 1 hour'::interval)) as t(interv);

create or replace function plluacu_ts_pl_span(a timestamp)
returns timestamp
language plluacu
as $$
  return timestamp.add_interval(a, {seconds= 10.4, minutes=4, hours= 5, days=5})
$$;
select '2008-01-01'::timestamp, plluacu_ts_pl_span('2008-01-01'::timestamp);

create or replace function plluacu_interval_norm(span_in interval)
returns setof text
language plluacu
as $$--[==[prepare
  local yield = spi.return_next
  local function out_table(name, t)
    for k, v in pairs(t) do
      yield(name .. '.' .. k .. '=' .. v)
    end
  end
--]==]
  out_table('span_in', span_in)
  local span_denorm = interval.denormalize(span_in)
  out_table('span_denorm', span_denorm)
  local span_out = interval.normalize(span_denorm)
  out_table('span_out', span_out)
$$;

select * from plluacu_interval_norm('1 year 3 months 34 days 3 hour 2 minute 1 second 4 milliseconds 5 microseconds'::interval);
select * from plluacu_interval_norm(-'1 year 3 months 34 days 3 hour 2 minute 1 second 4 milliseconds 5 microseconds'::interval);