set names 'sql_ascii';
set lc_messages='C';

\echo test_spi_lua1
create or replace function test_spi_lua1()
returns int
language plluacu
as $$
  local ret = spi.execute('select 1 as i')
  return ret[1].i;
$$;

select test_spi_lua1(); 

\echo test_spi_lua2
create or replace function test_spi_lua2()
returns setof varchar
language plluacu
as $$--[==[prepare
local yield = coroutine.yield
--]==]
  local ret = spi.execute('select * from test_lua')
  yield( #ret )
  for i, row in ipairs(ret) do
    yield("--- by position --")
    for k,v in ipairs(row) do
      yield( ("%s: %s"):format(k, tostring(v)))
    end
    yield("--- by name --")
    for k,v in spi.tuple_by_name(row) do
      yield( ("%s: %s"):format(k, tostring(v)))
    end
  end
$$;

select * from test_spi_lua2();

\echo test_spi_lua3
create or replace function test_spi_lua3(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.*, 1 from test_lua')
  return ret -- ignores fields with other names
$$;

select * from test_spi_lua3(1);

\echo test_spi_lua4
create or replace function test_spi_lua4(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.* from test_lua')
  return set{ ret } -- ignores fields with other names
$$;

select * from test_spi_lua4(1);

\echo test_spi_lua5
create or replace function test_spi_lua5(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.* from test_lua')
  return set( ret:fetch() )
$$;

select * from test_spi_lua5(1);

\echo test_spi_lua6
create or replace function test_spi_lua6(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.* from test_lua')
  return set{ ret:fetch() }
$$;

select * from test_spi_lua6(1);

\echo test_spi_lua7
create or replace function test_spi_lua7(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.execute('select test_lua.* from test_lua')
  return ret
$$;

select * from test_spi_lua7(1);

\echo test_spi_lua8
create or replace function test_spi_lua8(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.* from test_lua')
  return ret:iterate()
$$;

select * from test_spi_lua8(1);

\echo test_spi_lua9
create or replace function test_spi_lua9(i int)
returns setof test_lua
language plluacu
as $$
  local ret = spi.query('select test_lua.* from test_lua')
  for _, v in ret:iterate() do
    spi.return_next(v)
  end
$$;

select * from test_spi_lua9(1);

\echo test_spi_lua10
create or replace function test_spi_lua10(i int)
returns setof test_lua
language plluacu
as $$
  spi.return_next(set{ {name='ya', names={'yo','yu'}, ints={1,2,3,4} }, spi.query('select * from test_lua order by id')})
  spi.return_next(spi.query('select * from test_lua order by id desc'))
  return set{ {name='ya1', names={'yo1','yu1'}, ints={1,2,3,4} }, {name='ya2', names={'yo2','yu2'}, ints={1,2,3,4} } }
$$;

select * from test_spi_lua10(1);

\echo test_spi_lua11
create or replace function test_spi_lua11(i int4)
returns setof text[]
language plluacu
as $$
  COOL_QUERY = COOL_QUERY or spi.prepare([[
       select ARRAY['hi'|| $1,'ho'|| $1 ||j ]::varchar[] as ar from generate_series(1, 5) as j
       ]], 'int4')
  return COOL_QUERY:execute(i)
$$;

select * from test_spi_lua11(1) ;

\echo test_spi_lua12
create or replace function test_spi_lua12(i int4)
returns setof text[]
language plluacu
as $$--[==[prepare
  local COOL_QUERY = spi.prepare([[
       select ARRAY['hi'|| $1,'ho'|| $1 ||j ]::varchar[] as ar from generate_series(1, 5) as j
       ]], 'int4')
--]==]
  return COOL_QUERY:execute(i)
$$;
select * from test_spi_lua12(1);

\echo test_spi_lua13
create or replace function test_spi_lua13(i int4)
returns setof text[]
language plluacu
as $$
  COOL_QUERY1 = COOL_QUERY1 or spi.prepare([[
       select ARRAY['hi'|| $1,'ho'|| $1 ||j ]::varchar[] as ar from generate_series(1, 5) as j
       ]], 'int4')
  spi.return_next( COOL_QUERY1:query(i) )
$$;

select * from test_spi_lua13(1);

\echo test_spi_lua14
create or replace function test_spi_lua14(i int4)
returns setof float8
language plluacu
as $$
  COOL_QUERY4 = COOL_QUERY4 or spi.prepare([[
       select 0.1 + $1 as ar from generate_series(1, 5) as j
       ]], 'int4')
  return COOL_QUERY4:execute(spi.limit(4), i)
$$;

select * from test_spi_lua14(1);

\echo test_spi_lua15
create or replace function test_spi_lua15(i int4)
returns setof text
language plluacu
as $$
--  local yield = spi.return_next
  local yield = spi.info
  local function yield_hash(name, t)
    for k, v in spi.tuple_by_name(t) do
      yield("type( "..name.. "." .. k .. " ) = " .. type(v))
      yield(name.. "." .. k .. " = " .. tostring(v))
    end
  end
  local q = 'select * from test_lua order by id'
  local methods = { 
    {'execute', function(mode) return spi.execute(q, mode) end},
    {'fetch',   function(mode) return spi.query(q):fetch( nil, mode) end},
    {'prepare execute', function(mode) return spi.prepare(q):execute(mode) end},
    {'prepare fetch', function(mode) return spi.prepare(q):query():fetch( nil, mode) end}
    }
  local modes = {
    {'full', spi.d2l_mode.full},
    {'simple', spi.d2l_mode.simple},
    {'strings', spi.d2l_mode.strings}
  }
  for _, method in ipairs(methods) do
    local mtname, mtfunc = unpack(method)
    for _, mode in ipairs(modes) do
      local mdname, mdvalue = unpack(mode)
      yield("------- "..mtname.." "..mdname.." --------- ")
      for _, h in ipairs(mtfunc(mdvalue)) do
        yield_hash('r', h)
      end
    end
  end
$$;

select * from test_spi_lua15(1);
