set names 'sql_ascii';
set lc_messages='C';

drop table if exists test_lua_trig;
create table test_lua_trig (
  id serial primary key,
  name varchar,
  value text,
  amount numeric(8,2)
);

create or replace function test_lua_trig_1()
returns trigger
language plluacu
as $$--[==[prepare
  local function dump_table(tab, name)
   if tab then
    local out = {}
    for field, value in spi.tuple_by_name(tab) do
      table.insert(out, field..'= "'..tostring(value)..'"')
    end
    spi.notice('Trigger: '..name..'={' .. table.concat(out, ', ') ..'}')
   else
    spi.notice('Trigger: '..name..'==nil')
   end
  end
--]==]
  local args = {...}
  spi.notice('-------------------------------------')
  spi.notice('Trigger: ARGS: '..table.concat(args, ', '))
  dump_table(TD, 'TD')
--  for key, value in pairs(TD) do
--    spi.notice('Trigger: TD.'..key..'= "'..value..'"')
--  end
  dump_table(NEW, 'NEW')
  dump_table(OLD, 'OLD')
  local all = spi.execute('select * from test_lua_trig')
  spi.notice(('Trigger: existed: %d '):format(#all))
  for i, row in ipairs(all) do
    dump_table(row, 'row')
  end
$$;

create trigger test_lua_trig_ix_before_statement
before INSERT or DELETE or UPDATE
on test_lua_trig
for each statement execute procedure test_lua_trig_1('BEFORE', 'STATEMENT');

create trigger test_lua_trig_ix_after_statement 
after INSERT or DELETE or UPDATE
on test_lua_trig
for each statement execute procedure test_lua_trig_1('AFTER', 'STATEMENT');

create trigger test_lua_trig_ix_before_row
before INSERT or DELETE or UPDATE
on test_lua_trig
for each row execute procedure test_lua_trig_1('BEFORE', 'ROW');

create trigger test_lua_trig_ix_after_row
after INSERT or DELETE or UPDATE
on test_lua_trig
for each row execute procedure test_lua_trig_1('AFTER', 'ROW');

create or replace function test_lua_trig_2()
returns trigger
language plluacu
as $$
  spi.notice("UPDATE_TRIGGER for NEW.id = " .. NEW.id.. " !!!!!")
  local amount = (NEW.amount + OLD.amount)/2
  local value  = NEW.value..'!'
  spi.notice("UPDATE TRIGGER: set amount to "..amount)
  spi.notice("UPDATE TRIGGER: set value  to "..value)
  if NEW.id % 2 == 0 then
    return {amount= amount, value = value}
  else
    NEW.amount = amount
    NEW.value = value
    return NEW
  end
$$;

create trigger test_lua_trig_ix_xafter_row
before UPDATE
on test_lua_trig
for each row execute procedure test_lua_trig_2();


\echo inserting
insert into test_lua_trig(name, value, amount)
values ('name1', 'value1', 1),
       ('name2', 'value2', 2);

\echo updating
update test_lua_trig
set amount = amount + 0.5;

\echo deleting
delete from test_lua_trig;


drop table if exists test_lua_trig1 cascade;
create table test_lua_trig1
(
  id serial primary key,
  name varchar,
  fixed varchar
);
insert into test_lua_trig1(name, fixed) values ('hi', 'fixed');

\echo general_triggers_hard_way
create or replace function tr_general_fixed_fields()
returns trigger
language plluacu
as $$
--[==[prepare
  local eq_vals = spi.eq_values
--]==]
  local fixed_fields = { ... }
  local mapping = NEW.__mapping__
  for _, field_name in ipairs(fixed_fields) do
    local field_type = mapping[mapping[field_name]][2]
    if not eq_vals(NEW[field_name], OLD[field_name], field_type) then
      spi.error(
        ("Fixed field %q of table %q.%q has been tried to update"):format(
          field_name, TD.SCHEMA, TD.TABLE)
        )
    end
  end
$$;

create or replace function tr_general_skip_unchanged()
returns trigger
language plluacu
as $$
--[==[prepare
  local eq_vals = spi.eq_values
--]==]
  for _,v in ipairs(NEW.__mapping__) do
    local field_name, field_type = v[1], v[2]
    if not eq_vals(NEW[field_name], OLD[field_name], field_type) then
      return
    end
  end
  return NULL -- SKIP CHANGES
  --return 'SKIP'
$$;

create trigger tr_z_fixed before update
on test_lua_trig1
for each row execute procedure tr_general_fixed_fields('id','fixed');

create trigger tr_z_skip_unchaged before update
on test_lua_trig1
for each row execute procedure tr_general_skip_unchanged();

\echo run general_triggers_hard_way

update test_lua_trig1 set name = 'ho';
update test_lua_trig1 set name = 'ho';
update test_lua_trig1 set fixed = 'unfixed';

\echo general_triggers_simple_way

create or replace function tr_general_skip_unchanged()
returns trigger
language plluacu_simple
as $$
  for i, v in ipairs(NEW) do
    if v ~= OLD[i] then return end
  end
  return NULL -- SKIP CHANGES
$$;

create or replace function tr_general_fixed_fields()
returns trigger
language plluacu_simple
as $$
  local fixed_fields = { ... }
  for _, field_name in ipairs(fixed_fields) do
    if NEW[field_name] ~= OLD[field_name] then
      spi.error(
        ("Fixed field %q of table %q.%q has been tried to update"):format(
          field_name, TD.SCHEMA, TD.TABLE)
        )
    end
  end
$$;

\echo run general_triggers_simple_way

update test_lua_trig1 set name = 'hi';
update test_lua_trig1 set name = 'hi';
update test_lua_trig1 set fixed = 'unfixed';
